

Imports System.Diagnostics
Imports System.Runtime.InteropServices
Imports System.Threading
Imports LC.Threading
Imports LC.Multimedia

Public MustInherit Class OutputDeviceBase
    Inherits MidiDevice

    <DllImport("winmm.dll")> _
    Protected Shared Function midiOutReset(ByVal handle As Integer) As Integer
    End Function

    <DllImport("winmm.dll")> _
    Protected Shared Function midiOutShortMsg(ByVal handle As Integer, ByVal message As Integer) As Integer
    End Function

    <DllImport("winmm.dll")> _
    Protected Shared Function midiOutPrepareHeader(ByVal handle As Integer, ByVal headerPtr As IntPtr, ByVal sizeOfMidiHeader As Integer) As Integer
    End Function

    <DllImport("winmm.dll")> _
    Protected Shared Function midiOutUnprepareHeader(ByVal handle As Integer, ByVal headerPtr As IntPtr, ByVal sizeOfMidiHeader As Integer) As Integer
    End Function

    <DllImport("winmm.dll")> _
    Protected Shared Function midiOutLongMsg(ByVal handle As Integer, ByVal headerPtr As IntPtr, ByVal sizeOfMidiHeader As Integer) As Integer
    End Function

    <DllImport("winmm.dll")> _
    Protected Shared Function midiOutGetDevCaps(ByVal deviceID As Integer, ByRef caps As MidiOutCaps, ByVal sizeOfMidiOutCaps As Integer) As Integer
    End Function

    <DllImport("winmm.dll")> _
    Protected Shared Function midiOutGetNumDevs() As Integer
    End Function

    Protected Const MOM_OPEN As Integer = &H3C7
    Protected Const MOM_CLOSE As Integer = &H3C8
    Protected Const MOM_DONE As Integer = &H3C9

    Protected Delegate Sub GenericDelegate(Of T)(ByVal args As T)

    ' Represents the method that handles messages from Windows.
    Protected Delegate Sub MidiOutProc(ByVal handle As Integer, ByVal msg As Integer, ByVal instance As Integer, ByVal param1 As Integer, ByVal param2 As Integer)

    ' For releasing buffers.
    Protected delegateQueue As New DelegateQueue()

    Protected ReadOnly lockObject As New Object()

    ' The number of buffers still in the queue.
    Protected bufferCount As Integer = 0

    ' Builds MidiHeader structures for sending system exclusive messages.
    Private headerBuilder As New MidiHeaderBuilder()

    ' The device handle.
    Protected m_handle As Integer = 0

    Public Sub New(ByVal deviceID As Integer)
        MyBase.New(deviceID)
    End Sub

    Protected Overrides Sub Finalize()
        Try
            Dispose(False)
        Finally
            MyBase.Finalize()
        End Try
    End Sub

    Protected Overloads Overrides Sub Dispose(ByVal disposing As Boolean)
        If disposing Then
            delegateQueue.Dispose()
        End If
        MyBase.Dispose(disposing)
    End Sub

    Public Overridable Sub Send(ByVal message As ChannelMessage)
        If IsDisposed Then
            Throw New ObjectDisposedException(Me.GetType.Name)
        End If
        Send(message.Message)
    End Sub

    Public Overridable Sub Send(ByVal message As SysExMessage)
        If IsDisposed Then
            Throw New ObjectDisposedException(Me.[GetType]().Name)
        End If
        SyncLock lockObject
            headerBuilder.InitializeBuffer(message)
            headerBuilder.Build()

            ' Prepare system exclusive buffer.
            Dim result As Integer = midiOutPrepareHeader(m_handle, headerBuilder.Result, SizeOfMidiHeader)

            ' If the system exclusive buffer was prepared successfully.
            If result = MidiDeviceException.MMSYSERR_NOERROR Then
                bufferCount += 1

                ' Send system exclusive message.
                result = midiOutLongMsg(m_handle, headerBuilder.Result, SizeOfMidiHeader)

                ' If the system exclusive message could not be sent.
                If result <> MidiDeviceException.MMSYSERR_NOERROR Then
                    midiOutUnprepareHeader(m_handle, headerBuilder.Result, SizeOfMidiHeader)
                    bufferCount -= 1
                    headerBuilder.Destroy()

                    ' Throw an exception.
                    Throw New OutputDeviceException(result)
                End If
            Else
                ' Else the system exclusive buffer could not be prepared.
                ' Destroy system exclusive buffer.
                headerBuilder.Destroy()

                ' Throw an exception.
                Throw New OutputDeviceException(result)
            End If
        End SyncLock
    End Sub

    Public Overridable Sub Send(ByVal message As SysCommonMessage)
        If IsDisposed Then
            Throw New ObjectDisposedException(Me.GetType.Name)
        End If
        Send(message.Message)
    End Sub

    Public Overridable Sub Send(ByVal message As SysRealtimeMessage)
        If IsDisposed Then
            Throw New ObjectDisposedException(Me.GetType.Name)
        End If
        Send(message.Message)
    End Sub

    Public Overloads Overrides Sub Reset()
        If IsDisposed Then
            Throw New ObjectDisposedException(Me.GetType.Name)
        End If
        SyncLock lockObject
            ' Reset the OutputDevice.
            Dim result As Integer = midiOutReset(m_handle)

            If result = MidiDeviceException.MMSYSERR_NOERROR Then
                While bufferCount > 0
                    Monitor.Wait(lockObject)
                End While
            End If
        End SyncLock
    End Sub

    Protected Sub Send(ByVal message As Integer)
        SyncLock lockObject
            Dim result As Integer = midiOutShortMsg(m_handle, message)
        End SyncLock
    End Sub

    Public Shared Function GetDeviceCapabilities(ByVal deviceID As Integer) As MidiOutCaps
        Dim caps As New MidiOutCaps()

        ' Get the device's capabilities.
        Dim result As Integer = midiOutGetDevCaps(deviceID, caps, Marshal.SizeOf(caps))

        ' If the capabilities could not be retrieved.
        If result <> MidiDeviceException.MMSYSERR_NOERROR Then
            ' Throw an exception.
            Throw New OutputDeviceException(result)
        End If

        Return caps
    End Function

    Protected Overridable Sub HandleMessage(ByVal handle As Integer, ByVal msg As Integer, ByVal instance As Integer, ByVal param1 As Integer, ByVal param2 As Integer)
        If msg = MOM_OPEN Then
        ElseIf msg = MOM_CLOSE Then
        ElseIf msg = MOM_DONE Then
            delegateQueue.Post(AddressOf ReleaseBuffer, New IntPtr(param1))
        End If
    End Sub

    Private Sub ReleaseBuffer(ByVal state As Object)
        SyncLock lockObject
            Dim headerPtr As IntPtr = CType(state, IntPtr)

            ' Unprepare the buffer.
            Dim result As Integer = midiOutUnprepareHeader(m_handle, headerPtr, SizeOfMidiHeader)

            If result <> MidiDeviceException.MMSYSERR_NOERROR Then
                Dim ex As Exception = New OutputDeviceException(result)

                OnError(New ErrorEventArgs(ex))
            End If

            ' Release the buffer resources.
            headerBuilder.Destroy(headerPtr)

            bufferCount -= 1

            Monitor.Pulse(lockObject)

            Debug.Assert(bufferCount >= 0)
        End SyncLock
    End Sub

    Public Overloads Overrides Sub Dispose()
        If IsDisposed Then
            Return
        End If
        SyncLock lockObject
            Close()
        End SyncLock
    End Sub

    Public Overloads Overrides ReadOnly Property Handle() As Integer
        Get
            Return m_handle
        End Get
    End Property

    Public Shared ReadOnly Property DeviceCount() As Integer
        Get
            Return midiOutGetNumDevs()
        End Get
    End Property

    Public Overrides Sub Close()

    End Sub

End Class
