
Imports System.Diagnostics
Imports System.Collections.Generic
Imports System.ComponentModel

Public Class Sequencer
    Implements IComponent

    Private m_sequence As Sequence = Nothing
    Private enumerators As New List(Of IEnumerator(Of Integer))()
    Private dispatcher As New MessageDispatcher()
    Private chaser As New ChannelChaser()
    Private stopper As New ChannelStopper()
    Private clock As New MidiInternalClock()
    Private tracksPlayingCount As Integer
    Private ReadOnly lockObject As New Object()
    Private playing As Boolean = False
    Private m_disposed As Boolean = False
    Private m_site As ISite = Nothing


    ' ===============================================================================================
    ' "Events"
    ' ===============================================================================================

    Public Event PlayingCompleted As EventHandler

    Public Custom Event ChannelMessagePlayed As EventHandler(Of ChannelMessageEventArgs)
        AddHandler(ByVal value As EventHandler(Of ChannelMessageEventArgs))
            AddHandler dispatcher.ChannelMessageDispatched, value
        End AddHandler
        RemoveHandler(ByVal value As EventHandler(Of ChannelMessageEventArgs))
            RemoveHandler dispatcher.ChannelMessageDispatched, value
        End RemoveHandler
        RaiseEvent(ByVal sender As Object, ByVal value As ChannelMessageEventArgs)
        End RaiseEvent
    End Event

    Public Custom Event SysExMessagePlayed As EventHandler(Of SysExMessageEventArgs)
        AddHandler(ByVal value As EventHandler(Of SysExMessageEventArgs))
            AddHandler dispatcher.SysExMessageDispatched, value
        End AddHandler
        RemoveHandler(ByVal value As EventHandler(Of SysExMessageEventArgs))
            RemoveHandler dispatcher.SysExMessageDispatched, value
        End RemoveHandler
        RaiseEvent(ByVal sender As Object, ByVal value As SysExMessageEventArgs)
        End RaiseEvent
    End Event

    Public Custom Event MetaMessagePlayed As EventHandler(Of MetaMessageEventArgs)
        AddHandler(ByVal value As EventHandler(Of MetaMessageEventArgs))
            AddHandler dispatcher.MetaMessageDispatched, value
        End AddHandler
        RemoveHandler(ByVal value As EventHandler(Of MetaMessageEventArgs))
            RemoveHandler dispatcher.MetaMessageDispatched, value
        End RemoveHandler
        RaiseEvent(ByVal sender As Object, ByVal value As MetaMessageEventArgs)
        End RaiseEvent
    End Event

    Public Custom Event Chased As EventHandler(Of ChasedEventArgs)
        AddHandler(ByVal value As EventHandler(Of ChasedEventArgs))
            AddHandler chaser.Chased, value
        End AddHandler
        RemoveHandler(ByVal value As EventHandler(Of ChasedEventArgs))
            RemoveHandler chaser.Chased, value
        End RemoveHandler
        RaiseEvent(ByVal sender As Object, ByVal value As ChasedEventArgs)
        End RaiseEvent
    End Event

    Public Custom Event Stopped As EventHandler(Of StoppedEventArgs)
        AddHandler(ByVal value As EventHandler(Of StoppedEventArgs))
            AddHandler stopper.Stopped, value
        End AddHandler
        RemoveHandler(ByVal value As EventHandler(Of StoppedEventArgs))
            RemoveHandler stopper.Stopped, value
        End RemoveHandler
        RaiseEvent(ByVal sender As Object, ByVal value As StoppedEventArgs)
        End RaiseEvent
    End Event



    ' ===============================================================================================
    '  CONSTRUCTORS
    ' ===============================================================================================

    Public Sub New()
        AddHandler dispatcher.MetaMessageDispatched, AddressOf MetaMessageDispatchedDelegate
        AddHandler dispatcher.ChannelMessageDispatched, AddressOf ChannelMessageDispatchedDelegate
        AddHandler clock.Tick, AddressOf TickDelegate
    End Sub

    Private Sub MetaMessageDispatchedDelegate(ByVal sender As Object, ByVal e As MetaMessageEventArgs)
        If e.Message.MetaType = MetaType.EndOfTrack Then
            tracksPlayingCount -= 1
            If tracksPlayingCount = 0 Then
                Play_Stop()
                OnPlayingCompleted(EventArgs.Empty)
            End If
        Else
            clock.Process(e.Message)
        End If
    End Sub

    Private Sub ChannelMessageDispatchedDelegate(ByVal sender As Object, ByVal e As ChannelMessageEventArgs)
        stopper.Process(e.Message)
    End Sub

    Private Sub TickDelegate(ByVal sender As Object, ByVal e As EventArgs)
        SyncLock lockObject
            If Not playing Then
                Return
            End If
            For Each enumerator As IEnumerator(Of Integer) In enumerators
                enumerator.MoveNext()
            Next
        End SyncLock
    End Sub






    Protected Overrides Sub Finalize()
        Try
            Dispose(False)
        Finally
            MyBase.Finalize()
        End Try
    End Sub

    Protected Overridable Sub Dispose(ByVal disposing As Boolean)
        If disposing Then
            SyncLock lockObject
                Play_Stop()
                clock.Dispose()
                m_disposed = True
                GC.SuppressFinalize(Me)
            End SyncLock
        End If
    End Sub


    ' ===============================================================================================
    '   METHODS
    ' ===============================================================================================

    Public Sub Play_Start()
        If m_disposed Then
            Throw New ObjectDisposedException(Me.[GetType]().Name)
        End If
        SyncLock lockObject
            Play_Stop()
            Position = 0
            Play_Continue()
        End SyncLock
    End Sub

    Public Sub Play_Continue()
        If m_disposed Then
            Throw New ObjectDisposedException(Me.[GetType]().Name)
        End If
        If Sequence Is Nothing Then
            Return
        End If
        SyncLock lockObject
            Play_Stop()
            enumerators.Clear()

            For Each t As Track In Sequence
                Dim ti As Track.Tick_Iterator = New Track.Tick_Iterator(t, Position, chaser, dispatcher)
                enumerators.Add(ti.GetEnumerator)
            Next

            tracksPlayingCount = Sequence.Count

            playing = True
            clock.Ppqn = m_sequence.Division
            clock.ClockContinue()
        End SyncLock
    End Sub

    Public Sub Play_Stop()
        If m_disposed Then
            Throw New ObjectDisposedException(Me.[GetType]().Name)
        End If
        SyncLock lockObject
            If Not playing Then
                Return
            End If
            playing = False
            clock.ClockStop()
            stopper.AllSoundOff()
        End SyncLock
    End Sub

    Protected Overridable Sub OnPlayingCompleted(ByVal e As EventArgs)
        RaiseEvent PlayingCompleted(Me, e)
    End Sub

    Protected Overridable Sub OnDisposed(ByVal e As EventArgs)
        RaiseEvent Disposed(Me, e)
    End Sub

    Public Property Position() As Integer
        Get
            If m_disposed Then
                Throw New ObjectDisposedException(Me.[GetType]().Name)
            End If
            Return clock.Ticks
        End Get
        Set(ByVal value As Integer)
            If m_disposed Then
                Throw New ObjectDisposedException(Me.[GetType]().Name)
            ElseIf value < 0 Then
                Throw New ArgumentOutOfRangeException()
            End If

            Dim wasPlaying As Boolean

            SyncLock lockObject
                wasPlaying = playing
                Play_Stop()
                clock.SetTicks(value)
            End SyncLock

            SyncLock lockObject
                If wasPlaying Then
                    Play_Continue()
                End If
            End SyncLock
        End Set
    End Property



    Public ReadOnly Property PositionSeconds() As Double
        Get
            If m_disposed Then
                Throw New ObjectDisposedException(Me.[GetType]().Name)
            End If
            Dim beats As Double = clock.Ticks / clock.Ppqn
            Dim minutes As Double = beats / GetBPM
            Dim sec As Double = minutes * 60
            Return sec
        End Get
    End Property

    Public ReadOnly Property TotalSeconds() As Double

        ' 8         00:59 ---  33.03 ( 132 bps )
        ' 03_Per1   04:13
        ' 06_Emozi  05:19
        ' axel      02:30
        ' blues1    06:42
        ' dance     01.34
        ' Englishm. 04:15

        ' Bwild     03:12
        ' Buebyes   02:28
        ' CloneLove 02:15


        Get
            If m_disposed Then
                Throw New ObjectDisposedException(Me.[GetType]().Name)
            End If

            '                       03_PER01    8.mid       axel        blues1
            ' -----------------------------------------------------------------------
            ' time                  04:13       00:59       02:30       06:42
            ' seconds               253         59          150         402
            ' ---------------------------------------------------------------------------- clock
            '  .seqType             0(ppqn)     0(ppqn)     0(ppqn)     0(ppqn)
            '  .DefaultTempo        50000       50000       50000       50000
            '  .Tempo               645161      454545      500000      1090909

            '  .Ppqn                96          480         192         480
            '  .periodResolution    96000       480000      192000      480000
            '  .PpqnMinValue        24          24          24          24
            '  .TicksPerClock       4           20          8           20
            ' ---------------------------------------------------------------------------- m_sequence
            '  .division            96          480         192         480       

            '  .GetLength           38017       61391       262144      178561

            ' BPM                   93          132         120         55 

            ' .GetLength / sec      150.265     1040.53     1747.627    444.18

            'Dim sec As Double = m_sequence.GetLength / 2.0 / clock.Ppqn
            'Dim sec As Double = m_sequence.GetLength * 60000.0 * 4.0 / clock.Tempo / clock.Ppqn

            Dim beats As Double = m_sequence.GetLength / clock.Ppqn
            Dim minutes As Double = beats / GetBPM
            Dim sec As Double = minutes * 60
            Return sec
        End Get
    End Property

    Public ReadOnly Property GetBPM() As Double
        Get
            If m_disposed Then
                Throw New ObjectDisposedException(Me.[GetType]().Name)
            End If
            Return (1000000.0 * 60.0) / clock.Tempo
        End Get
    End Property


    Public Property Sequence() As Sequence
        Get
            Return m_sequence
        End Get
        Set(ByVal value As Sequence)
            If value Is Nothing Then
                Throw New ArgumentNullException()
            ElseIf value.SequenceType = SequenceType.Smpte Then
                Throw New NotSupportedException()
            End If

            SyncLock lockObject
                Play_Stop()
                m_sequence = value
            End SyncLock
        End Set
    End Property

#Region "IComponent Members"

    Public Event Disposed As EventHandler Implements System.ComponentModel.IComponent.Disposed

    Public Property Site() As ISite Implements IComponent.Site
        Get
            Return m_site
        End Get
        Set(ByVal value As ISite)
            m_site = value
        End Set
    End Property

#End Region

#Region "IDisposable Members"

    Public Sub Dispose() Implements IDisposable.Dispose
        If m_disposed Then
            Return
        End If

        Dispose(True)
    End Sub

#End Region

End Class
